/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#define DBG 1
#define DBG_ROUND 0
#define DBG_RDKEY 1
#else
#define DBG 0
#define DBG_ROUND 0
#define DBG_RDKEY 0
#endif

#include "ap_int.h"
#include "hls_stream.h"
using namespace hls;
#include "hls_interface.h"
#include "hls_utils.h"
#include "hls_aes.h"
#include "hls_deoxys.h"

enum states_t {S_INIT, S_AD, S_MSG_AD_EMPTY, S_MSG, S_TAG_INIT_PARTIAL, S_TAG_INIT, S_TAG};

#define ROUNDNR         14
#define CAESAR_R3       1

#define MSB_AD 			      (0x2<<4)
#define MSB_AD_LAST		      (0x6<<4)
#define MSB_M			      (0x0<<4)
#define MSB_M_LAST_NONZERO	  (0x4<<4)
#define MSB_CHKSUM_FULL       (0x1<<4)
#define MSB_CHKSUM_NON_FULL	  (0x5<<4)

/* Number of bits in the TWEAKEY state (256 or 384) */
#define TWEAKEY_STATE_SIZE    256

ap_uint<8> xtime2(ap_uint<8> a);

#if DBG
static void printState(ap_uint<8> *tweak, int size) {
    int i;
    for (i=0; i<size; i++) {
        std::cout << std::hex << std::uppercase << std::setw(2) << std::setfill('0')
                  << (int) tweak[i];
        if ((i+1)%8 == 0) std::cout << " ";
    }
    std::cout << std::endl;
}
#endif

static const ap_uint<8> rcon_rom[16] = {
    0x5E,0xBC,0x63,0xC6,
    0x97,0x35,0x6A,0xD4,
    0xB3,0x7D,0xFA,0xEF,
    0xC5,0x91,0x39,0x72
};


/**********************************************************************************
*** In Deoxys=/=-128-128, the tweak is on 128 bits:
***     tweak = <stage> || <nonce> || <blockNumber>
***  where we use:
***      3 bits for stage
***     64 bits for nonce
***     61 bits for blockNumber
***********************************************************************************/

static void set_tweak(ap_uint<8> *tweak, const ap_uint<8> nonce[8], 
                      const ap_uint<32> block_no, const ap_uint<8> value)
{
    tweak[0] = (nonce[0] >> 4) ^ value;
    tweak[1] = (nonce[0]&0xf) << 4 ^ (nonce[1] >> 4);
    tweak[2] = (nonce[1]&0xf) << 4 ^ (nonce[2] >> 4);
    tweak[3] = (nonce[2]&0xf) << 4 ^ (nonce[3] >> 4);
    tweak[4] = (nonce[3]&0xf) << 4 ^ (nonce[4] >> 4);
    tweak[5] = (nonce[4]&0xf) << 4 ^ (nonce[5] >> 4);
    tweak[6] = (nonce[5]&0xf) << 4 ^ (nonce[6] >> 4);
    tweak[7] = (nonce[6]&0xf) << 4 ^ (nonce[7] >> 4);
    tweak[8] = (nonce[7]&0xf) << 4;
    tweak[ 9] = 0;
    tweak[10] = 0;
    tweak[11] = 0;
    tweak[12] = ((block_no >> 24ULL) & 0xff);
    tweak[13] = ((block_no >> 16ULL) & 0xff);
    tweak[14] = ((block_no >>  8ULL) & 0xff);
    tweak[15] = ((block_no >>  0ULL) & 0xff);
}

/*
** Modify the nonce part in the tweak value
*/
static void set_nonce_in_tweak(ap_uint<8> *tweak, const ap_uint<8> *nonce) {
#pragma HLS INLINE
    tweak[0] = nonce[0] >> 4;
    tweak[1] = (nonce[0]&0xf) << 4 ^ (nonce[1] >> 4);
    tweak[2] = (nonce[1]&0xf) << 4 ^ (nonce[2] >> 4);
    tweak[3] = (nonce[2]&0xf) << 4 ^ (nonce[3] >> 4);
    tweak[4] = (nonce[3]&0xf) << 4 ^ (nonce[4] >> 4);
    tweak[5] = (nonce[4]&0xf) << 4 ^ (nonce[5] >> 4);
    tweak[6] = (nonce[5]&0xf) << 4 ^ (nonce[6] >> 4);
    tweak[7] = (nonce[6]&0xf) << 4 ^ (nonce[7] >> 4);
    tweak[8] = (nonce[7]&0xf) << 4;
#if DBG
    std::cout << "Tweak - after set nonce\t: "; printState(tweak, 16);
#endif
}


/*
** Modify the block number in the tweak value
*/
static void set_block_number_in_tweak(ap_uint<8> *tweak, const ap_uint<32> block_no) {
#pragma HLS INLINE
    // tweak[ 8] = (tweak[8]&0xf0) ^ ((block_no >> 56ULL) & 0xf);
    // tweak[ 9] = ((block_no >> 48ULL) & 0xff);
    // tweak[10] = ((block_no >> 40ULL) & 0xff);
    // tweak[11] = ((block_no >> 32ULL) & 0xff);

    /* Note: The top 4 words are zeros as block_no can never
             reach that high based on our assumption */
    tweak[ 8] = tweak[8]&0xf0;
    tweak[ 9] = 0;
    tweak[10] = 0;
    tweak[11] = 0;
    tweak[12] = ((block_no >> 24ULL) & 0xff);
    tweak[13] = ((block_no >> 16ULL) & 0xff);
    tweak[14] = ((block_no >>  8ULL) & 0xff);
    tweak[15] = ((block_no >>  0ULL) & 0xff);
#if DBG
    std::cout << "Tweak - after set block\t: "; printState(tweak, 16);
#endif
}

/*
** Modify the stage value in the tweak value
*/
static void set_stage_in_tweak(ap_uint<8> *tweak, const ap_uint<8> value) {
#pragma HLS INLINE
    tweak[0]=(tweak[0] & 0xf) ^ value ;
#if DBG
    std::cout << "Tweak - after set stage\t: "; printState(tweak, 16);
#endif
}

/*
** Update the tweak value in the tweakey word.
** In the case of Deoxys-BC-256, the tweakey word is composed of KEY || TWEAK.
** In the case of Deoxys-BC-384, the tweakey word is composed of KEY_2 || KEY_1 || TWEAK.
*/
static void set_tweak_in_tweakey(ap_uint<8> *tweakey, ap_uint<8> *tweak) {
#pragma HLS INLINE
#if TWEAKEY_STATE_SIZE==256
    copy(tweakey+16, tweak);
#elif TWEAKEY_STATE_SIZE==384
    copy(tweakey+32, tweak);
#endif

#if DBG
    std::cout << "Tweakey init \t\t: "; printState(tweakey, 32);
#endif
}


/* Tweakey expansion */
static void GH(ap_uint<8> *tweakey) {
#pragma HLS INLINE
    size_t    i, j;
    ap_uint<8>  h[32];
    ap_uint<8>  g[32];
#pragma HLS ARRAY_RESHAPE variable=h     complete dim=1
#pragma HLS ARRAY_RESHAPE variable=g     complete dim=1

    static const ap_uint<8> perm[16] = {1,6,11,12,5,10,15,0,9,14,3,4,13,2,7,8};
    // H
    for (i = 0; i < 2; i++) {
#pragma HLS UNROLL
        for (j = 0; j < 16; j++) {
#pragma HLS UNROLL
            h[16*i+perm[j]] = tweakey[16*i+j];
        }
    }

    // G
#if CAESAR_R3
#pragma HLS RESOURCE variable=lfsr2_00 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_01 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_02 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_03 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_04 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_05 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_06 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_07 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_08 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_09 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_10 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_11 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_12 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_13 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_14 core=ROM_1P_1S
#pragma HLS RESOURCE variable=lfsr2_15 core=ROM_1P_1S
    g[ 0] =  lfsr2_00[h[ 0]];
    g[ 1] =  lfsr2_01[h[ 1]];
    g[ 2] =  lfsr2_02[h[ 2]];
    g[ 3] =  lfsr2_03[h[ 3]];
    g[ 4] =  lfsr2_04[h[ 4]];
    g[ 5] =  lfsr2_05[h[ 5]];
    g[ 6] =  lfsr2_06[h[ 6]];
    g[ 7] =  lfsr2_07[h[ 7]];
    g[ 8] =  lfsr2_08[h[ 8]];
    g[ 9] =  lfsr2_09[h[ 9]];
    g[10] =  lfsr2_10[h[10]];
    g[11] =  lfsr2_11[h[11]];
    g[12] =  lfsr2_12[h[12]];
    g[13] =  lfsr2_13[h[13]];
    g[14] =  lfsr2_14[h[14]];
    g[15] =  lfsr2_15[h[15]];
#endif
    for (i = 0; i < 16; i++) {
    #pragma HLS UNROLL
#if !CAESAR_R3
        g[   i] = xtime2(h[i]);
#endif
        g[16+i] =        h[16+i];
    }

#if 0
    std::cout << "\tH : "; printState(h, 32);
    std::cout << "\tG : "; printState(g, 32);
#endif

    // Save
    copy(tweakey, g);
    copy(tweakey+16, g+16);
}

void hls_ciphercore (
    BDO_STREAM &bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM &bdi_bus,
    KEY_STREAM &key_bus,
#if MANUAL_RAM 
    volatile ap_uint<BLOCKSIZE> *ramdi1, 
    volatile ap_uint<BLOCKSIZE> *ramdo1, 
    volatile ap_uint<4> *ramaddr1,
    volatile ap_uint<BLOCKSIZE> *ramdi2, 
    volatile ap_uint<BLOCKSIZE> *ramdo2, 
    volatile ap_uint<4> *ramaddr2,
#elif __SYNTHESIS__
    ap_uint<128> rdkey1[16],
    ap_uint<128> rdkey2[16],
#endif
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE port=msg_auth_valid axis
#if MANUAL_RAM
#pragma HLS INTERFACE ap_vld  port=ramdi1
#pragma HLS INTERFACE ap_none port=ramdo1
#pragma HLS INTERFACE ap_vld  port=ramdi2
#pragma HLS INTERFACE ap_none port=ramdo2
#pragma HLS INTERFACE port=ramaddr1 axis
#pragma HLS INTERFACE port=ramaddr2 axis
#else
    #if !__SYNTHESIS__
    ap_uint<128> rdkey1[16];
    ap_uint<128> rdkey2[16];
    #endif
#endif
    /* Resource declaration */
#pragma HLS RESOURCE variable=rcon_rom  core=ROM_1P_1S
    /* Interface related temporary variables */
    public_bus bdi_t1;
    static public_bus bdi_r;
    output_bus bdo;
    secret_bus key;
    array_t bdo_data;
    array_t bdi_data_t1;

#pragma HLS ARRAY_RESHAPE variable=bdo_data complete dim=1
#pragma HLS ARRAY_RESHAPE variable=bdi_data_t1 complete dim=1

    /* variables */
    static ap_uint<KEYSIZE> key_reg;
    states_t state = S_INIT;
    array_t checksum;
    array_t auth;
    array_t s;
    static ap_uint<8> npub[8];
    static ap_uint<8> tweakey[32];
    array_t rk;
    static array_t rk_init;

#pragma HLS ARRAY_RESHAPE variable=tweakey  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=checksum complete dim=1
#pragma HLS ARRAY_RESHAPE variable=auth     complete dim=1
#pragma HLS ARRAY_RESHAPE variable=s        complete dim=1
#pragma HLS ARRAY_RESHAPE variable=npub     complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk       complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk_init  complete dim=1

    /* Temporary variables */
    const ap_uint<8> ZEROS[8] = {0,0,0,0,0,0,0,0};
    array_t          new_rk;
    array_t          new_s;
    array_t          tag;
    ap_uint<128>     ram_data1;
    ap_uint<128>     ram_data2;

#pragma HLS ARRAY_RESHAPE variable=ZEROS    complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_rk  complete dim=1
#pragma HLS ARRAY_RESHAPE variable=new_s   complete dim=1
#pragma HLS ARRAY_RESHAPE variable=tag    complete dim=1            

    size_t i, j;
    ap_uint<4> addr1;
    ap_uint<4> addr2;
    bool op_decrypt = 0;
    bool op_decrypt_r = 0;
    bool sel_wr_ram1 = 0;
    bool is_partial = 0;
    bool sel_npub   = 0;

    /* Initialization */
    size_t loop = 1;
    static ap_uint<MAX_LEN-7> block_no;
    block_no = 0;

    bdo.size = 0;
#if !__SYNTHESIS__
    size_t dbg_loop = 0;
#endif

    // Key initialization (if needed)
    if (key_update == 1) {
#if KEYBUS == 128
        key_bus.read(key); key_reg = key.data;
#else
        LOAD_KEY: for (i=0; i<4; i++) {
            key_bus.read(key);
            key_reg = (key_reg << KEYBUS) ^ key.data;
        }
#endif
    }
    clear(auth);
    clear(checksum);

    // Read IV
    bdi_bus.read(bdi_t1);
#if DBG
    std::cout << "... reading npub\n" << std::endl;
#endif

    reshape_block(bdi_data_t1, bdi_t1.data);
    for (i=0; i<8; i++) {
    #pragma HLS UNROLL
        npub[i] = bdi_data_t1[i];
    }
    if (bdi_t1.eoi) state = S_TAG_INIT;
    bdi_r.data = 0;
    bdi_r.type = 0;
    bdi_r.valid_bytes = 0;
    bdi_r.pad_loc = 0;
    bdi_r.eot = 0;
    bdi_r.eoi = 0;
#pragma HLS DATA_PACK variable=bdi_r

    MSG_LOOP:
    while (loop) {
#if !__SYNTHESIS__
        // Add this clause to prevent infinite loop during debugging
        if (dbg_loop > 300) return;
        else dbg_loop++;
#endif
#if DBG
        std::cout << "Current state = " << state << std::endl;
#endif

        // PreProcessing
        array_t input;
        array_t bdi_data;
        bool is_auth, is_data;
        bool is_full, is_eoi, is_eot;
        ap_uint<3> is_type;
        bool new_data;
        data_bytes_t     bdi_valid_bytes;
        data_bytes_t     bdi_pad_loc;
        ap_uint<8>       stage_status;
        ap_uint<8>       stage_status_tmp;
#pragma HLS ARRAY_RESHAPE variable=input complete dim=1
#pragma HLS ARRAY_RESHAPE variable=bdi_data complete dim=1
        public_bus bdi;

        /* Note: This approach produces an additional clock cycle */
        if ((state == S_MSG || state == S_AD) && !is_partial) {
            new_data = 1;
            bdi_bus.read(bdi);
#if DBG
            std::cout << "... reading" << std::endl;
#endif
        } else {
            new_data = 0;
            bdi = bdi_r;
        }
        reshape_block(bdi_data, bdi.data);
        bdi_valid_bytes = bdi.valid_bytes;
        bdi_pad_loc = bdi.pad_loc;
        is_type = bdi.type;
        is_full = bdi.valid_bytes & 0x1;
        is_eot = bdi.eot;
        is_eoi = bdi.eoi;

        if ((state == S_MSG && !is_partial)
            || (state == S_AD)
            || (state == S_MSG_AD_EMPTY)) {
            copy(input, bdi_data);
        } else if (state == S_TAG) {
            copy(input, checksum);
        } else {
            clear(input);
        }
        
        ap_uint<MAX_LEN-7> block_no_new = 0;
        switch (state) {
        case S_INIT:
            block_no_new = 0;
            break;
        case S_AD:
            if (!is_partial) {
                if (is_type == TYPE_AD) {
                    if (is_eot) {
                        if (is_full) {
                            block_no_new = 0;
                        } else {
                            block_no_new = block_no;
                        }
                    } else {
                        block_no_new = block_no + 1;
                    }
                } else {
                    block_no_new = block_no;
                }
            } else {
                block_no_new = 0;
            }
            break;
        case S_MSG_AD_EMPTY:
            block_no_new = block_no + is_full;
            break;
        case S_MSG:
            if (!is_partial) {
                block_no_new = block_no + is_full;
            } else {
                block_no_new = block_no;
            }
            break;
        case S_TAG_INIT: 
            block_no_new = block_no;
            break;
        case S_TAG_INIT_PARTIAL:
        default: // S_TAG
            block_no_new = block_no + 1;
        }

        /* Adds a temporary variable here to reduce the critical path.
         * The tool adds an additional clock cycle otherwise */

        if (is_eot) {
            if (is_full) {
                stage_status_tmp = is_eoi ? MSB_CHKSUM_FULL : MSB_M;
            } else {
                stage_status_tmp = MSB_AD_LAST;
            }
        } else {
            stage_status_tmp = MSB_AD;
        }

        /* 'if' based stage_status selection */

        if (state == S_INIT) {
            stage_status = MSB_AD;
        } else if (state == S_AD) {
            if (!is_partial) {
                if (is_type == TYPE_AD) {
                    stage_status = stage_status_tmp;
                } else {
                    stage_status = is_full ? MSB_M : MSB_M_LAST_NONZERO;
                }
            } else {
                stage_status = is_eoi ? MSB_CHKSUM_FULL : MSB_M;
            }
        } else if (state == S_MSG_AD_EMPTY) {
            if (is_eot) 
                stage_status = is_full ? MSB_CHKSUM_FULL : MSB_M_LAST_NONZERO;
            else
                stage_status = MSB_M;
        } else if (state == S_MSG) {
            if (!is_partial) {
                if (is_eot) {
                    stage_status = is_full ? MSB_CHKSUM_FULL : MSB_M_LAST_NONZERO;
                } else {
                    stage_status = MSB_M;
                }
            } else {
                stage_status = MSB_CHKSUM_NON_FULL;
            }
        } else if (state == S_TAG_INIT_PARTIAL) {
            stage_status = MSB_CHKSUM_NON_FULL;
        } else if (state == S_TAG_INIT) {
            stage_status = MSB_CHKSUM_FULL;
        } else {
            stage_status = 0;
        }

#if 0
        /* Note: Keep this here for later reference 
         * The below code, based on switch, adds an additional clock cycle.
         * The if statement in the above doesn't */

        switch(state) {
        case S_INIT:
            stage_status = MSB_AD;
            break;
        case S_AD:
            if (!is_partial) {
                if (is_type == TYPE_AD) {
                    stage_status = stage_status_tmp;
                } else {
                    stage_status = is_full ? MSB_M : MSB_M_LAST_NONZERO;
                }
            } else {
                stage_status = is_eoi ? MSB_CHKSUM_FULL : MSB_M;
            }
            break;
        case S_MSG_AD_EMPTY:
            if (is_eot) 
                stage_status = is_full ? MSB_CHKSUM_FULL : MSB_M_LAST_NONZERO;
            else
                stage_status = MSB_M;
            break;
        case S_MSG:
            if (!is_partial) {
                if (is_eot) {
                    stage_status = is_full ? MSB_CHKSUM_FULL : MSB_M_LAST_NONZERO;
                } else {
                    stage_status = MSB_M;
                }
            } else {
                stage_status = MSB_CHKSUM_NON_FULL;
            }
            break;
        case S_TAG_INIT_PARTIAL:
            stage_status = MSB_CHKSUM_NON_FULL;
            break;
        case S_TAG_INIT: 
            stage_status = MSB_CHKSUM_FULL;
            break;
        default: // S_TAG
            stage_status = 0;
            break;
        }
#endif

        if ((state == S_INIT) 
            || (state == S_AD && !is_partial && is_type == TYPE_AD && (!is_eot || !is_full))) {
            sel_npub = 1;
        } else {
            sel_npub = 0;
        }

        /* Initial Add round key */
        // Tweak setup
        reshape_block(tweakey, key_reg);
#ifdef MORE_RESOURCE
        // This construct needs more resource
        if (sel_npub)
            set_nonce_in_tweak(tweakey+16, ZEROS);
        else
            set_nonce_in_tweak(tweakey+16, npub);
        set_stage_in_tweak(tweakey+16, stage_status);
        set_block_number_in_tweak(tweakey+16, (ap_uint<32>) block_no_new);
#else
        ap_uint<8> npub_or_zero[8];
        for (i=0; i<8; i++) {
        #pragma HLS UNROLL
            npub_or_zero[i] = (sel_npub) ? (ap_uint<8>) 0 : npub[i];
        }
        set_tweak(tweakey+16, npub_or_zero, block_no_new, stage_status);
#endif
        block_no = block_no_new;

#if DBG
        std::cout << "Tweakey init   : "; printState(tweakey, 32);
        //std::cout << "rcon : "; printState(rcon, 16);
        //std::cout << "npub : " << sel_npub << std::endl;
        std::cout << "  Block number\t: " << (int) block_no << std::endl;
        std::cout << "  Input \t: "; printState(input, 16);
        std::cout << "  init key\t: "; printState(rk_init, 16);
        std::cout << "  Auth\t\t: "; printState(auth, 16);
        std::cout << "  Checksum\t: "; printState(checksum, 16);
#endif

        /* New block operation */
        array_t  rcon2;
        rcon2[0] = 0x01;
        rcon2[1] = 0x02;
        rcon2[2] = 0x04;
        rcon2[3] = 0x08;
        for (size_t j=0; j<4; j++) {
        #pragma HLS UNROLL
            rcon2[4+j] = 0x2F;
        }
        for (size_t j=8; j<16; j++) {
        #pragma HLS UNROLL
            rcon2[j] = 0;
        }
        xor2(s, input, rk_init);
        xor3(new_rk, tweakey, tweakey+16, rcon2);
#if DBG_RDKEY
        std::cout << "\tNew Round key\t: "; printState(new_rk, 16);
        std::cout << "\tWrite ram 1 ?" << sel_wr_ram1 << std::endl;
        for (size_t j=0; j<ROUNDNR; j++) {
            std::cout << "\tRK   (" << (int) j << "): " << (sel_wr_ram1 ? rdkey2[j]:rdkey1[j]) << std::endl;
        }
#endif
        copy(rk_init, new_rk);

#if MANUAL_RAM
        *ramaddr1 = ROUNDNR-1;
        *ramaddr2 = ROUNDNR-1;
        if (sel_wr_ram1) {
            *ramdi1 = reshape_array(new_rk);
        } else {
            *ramdi2 = reshape_array(new_rk);
        }
#else
        if (sel_wr_ram1)
            rdkey1[ROUNDNR-1] = reshape_array(new_rk);
        else
            rdkey2[ROUNDNR-1] = reshape_array(new_rk);
#endif

#if DBG
        std::cout << "\tInput xor Roundkey\t: "; printState(s, 16);
#endif

        switch(state) {
        case S_AD:
            if (!is_partial) {
                if (is_type == TYPE_AD) {
                    if (is_eot && is_full)
                        op_decrypt = is_eoi ? 0 : decrypt;
                    else
                        op_decrypt = 0;
                } else {
                    op_decrypt = decrypt;
                }
            } else {
                op_decrypt = is_eoi ? 0 : decrypt;
            }
            break;
        case S_MSG_AD_EMPTY:
            op_decrypt = (is_eot) ? 0 : decrypt;
            break;
        case S_MSG:
            if (!is_partial) {
                op_decrypt = (is_eot) ? 0 : decrypt;
            } else {
                op_decrypt = 0;
            }
            break;
        default:
            op_decrypt = 0;
            break;
        }

        /* Register region */

        if (state == S_MSG_AD_EMPTY) {
            is_data = is_full;
        } else if (state == S_MSG) {
            is_data = (is_partial) ? 1 : is_full;
        } else {
            is_data = 0;
        }
        
        if (state == S_AD) {
            if (is_partial) {
                is_auth = 1;
            } else {
                is_auth = (is_type == TYPE_AD) ? is_full : 0;
            }
        } else {
            is_auth = 0;
        }

        if (new_data) {
            bdi_r = bdi;
        }

        /* Main operation */
        ap_uint<4> invi = ROUNDNR-2;
        RoundLoop:
        for (i=0; i<ROUNDNR; i++) {
            /* Tweakey update, new sub-round key generation,
             * and loading of sub-roundkey 
             **/
            ap_uint<8> rcon_rnd = rcon_rom[i];
            array_t  rcon;
            rcon[0] = 0x01;
            rcon[1] = 0x02;
            rcon[2] = 0x04;
            rcon[3] = 0x08;
            for (size_t j=0; j<4; j++) {
            #pragma HLS UNROLL
                rcon[4+j] = rcon_rnd;
            }
            for (size_t j=8; j<16; j++) {
            #pragma HLS UNROLL
                rcon[j] = 0;
            }

            GH(tweakey);
            xor3(new_rk, tweakey, tweakey+16, rcon);

            if (sel_wr_ram1) {
                addr1 = op_decrypt ? invi : (ap_uint<4>) i;
                addr2 = i;
            } else {
                addr1 = i;
                addr2 = op_decrypt ? invi : (ap_uint<4>) i;
            }

            
#if MANUAL_RAM
            *ramaddr1 = addr1;
            *ramaddr2 = addr2;
            ram_data1 = *ramdo1;
            ram_data2 = *ramdo2;
            if (sel_wr_ram1) {
                *ramdi1 = reshape_array(new_rk);
            } else {
                *ramdi2 = reshape_array(new_rk);
            }
#else
            ram_data1 = rdkey1[addr1];
            ram_data2 = rdkey2[addr2];
            if (sel_wr_ram1)
                rdkey1[addr1] = reshape_array(new_rk);
            else
                rdkey2[addr2] = reshape_array(new_rk);
#endif
            if (sel_wr_ram1) {
                reshape_block(rk, ram_data2);
            } else {
                reshape_block(rk, ram_data1);
            }

            /* Main round */
            if (op_decrypt_r) {
                AESInvRoundBasic(new_s, s, rk);
            } else {
                AESRoundBasic(new_s, s, rk);
            }
#if DBG_ROUND
            std::cout << "\t-- Round(" << (int) i << ")" << std::endl;
            std::cout << "\tState: "; printState(s, 16);
            std::cout << "\tRK   : "; printState(rk, 16);
            std::cout << "\tnewS : "; printState(new_s, 16);
#endif
            /* Update round key if preparing for decryption */
            if (op_decrypt)
                copy(rk_init, new_rk);
            copy(s, new_s);
            invi--;
        }
#if DBG
        std::cout << "\tStateOut\t\t: "; printState(s, 16);
        std::cout << "\top_dec     : " << (int) op_decrypt << std::endl;
        std::cout << "\top_dec_r   : " << (int) op_decrypt_r << std::endl;
        std::cout << "\twrite_ram1?: " << (int) sel_wr_ram1 << std::endl;
        std::cout << "\tdecrypt    : " << (int) decrypt << std::endl;
        std::cout << "\tis_type    : " << (int) is_type << std::endl;
        std::cout << "\tis_eoi     : " << (int) is_eoi << std::endl;
        std::cout << "\tis_eot     : " << (int) is_eot << std::endl;
        std::cout << "\tis_full    : " << (int) is_full << std::endl;
        std::cout << "\tis_auth    : " << (int) is_auth << std::endl;
        std::cout << "\tis_data    : " << (int) is_data << std::endl;
        std::cout << "\t(cur) is_partial : " << (int) is_partial << std::endl;
#endif

        sel_wr_ram1 = !sel_wr_ram1;
        op_decrypt_r = op_decrypt;

        // Post Processing
        array_t tmp_out;
        array_t message;
        array_t output;
        array_t data_r;
        #pragma HLS ARRAY_RESHAPE variable=tmp_out complete dim=1
        #pragma HLS ARRAY_RESHAPE variable=output complete dim=1
        #pragma HLS ARRAY_RESHAPE variable=message complete dim=1
        #pragma HLS ARRAY_RESHAPE variable=data_r complete dim=1

        reshape_block(data_r, bdi_r.data);
        for (size_t j=0; j<ELEMENTS; j++) {
        #pragma HLS UNROLL
            ap_uint<8> tmploc = ((ap_uint<8>) ((bdi_r.pad_loc >> (ELEMENTS-j-1)) & 0x1)) << 7;
            if ((bdi_r.valid_bytes >> ELEMENTS-1-j) & 0x1)
                tmp_out[j] = (s[j] ^ data_r[j]) ^ tmploc;
            else
                tmp_out[j] = tmploc;
        }
        if (decrypt) 
            if (is_full)
                copy(message, s);
            else
                copy(message, tmp_out);
        else
            copy(message, data_r);
        xor2(tag, s, auth);

        if (is_partial)
            xor2(output, s, data_r);
        else
            copy(output, s);

#if DBG
        std::cout << "\ttmp_out : "; printState(tmp_out, 16);
        std::cout << "\ttag     : "; printState(tag, 16);
#endif

        /* Update Auth */
        if (is_auth)
            xor2(auth, auth, s);

        /* Output Tag */
        if (state == S_TAG) {
            if (decrypt) {
                array_t data_t;
                #pragma HLS ARRAY_RESHAPE variable=data_t complete dim=1
                bdi_bus.read(bdi); reshape_block(data_t, bdi.data);
#if DBG
                std::cout << "... reading tag\n" << std::endl;
#endif
                *msg_auth_valid = match(data_t, tag);
            } else {
                bdo.data = reshape_array(tag);
                bdo_bus.write(bdo);
#if DBG
                std::cout << "... writing tag\n" << std::endl;
#endif
            }
            return;
        } else {
            /* Update checksum and output data */
            if (is_data) {
                xor2(checksum, checksum, message);
                bdo.data = reshape_array(output); bdo_bus.write(bdo);
#if DBG
                std::cout << "... writing\n" << std::endl;
#endif
            }
        }

        //if (new_data) 
        //    bdi_r = bdi;

        switch(state) {
        case S_INIT: state = S_AD; break;
        case S_AD: 
            if (!is_partial) {
                if (is_type == TYPE_AD) {
                    if (is_eot && is_full)
                        state = is_eoi ? S_TAG : S_MSG;
                    if (is_eot && !is_full)
                        is_partial = 1;
                } else {
                    state = S_MSG_AD_EMPTY;
                }
            } else {
                state = is_eoi ? S_TAG : S_MSG;
                is_partial = 0;
            }
            break;
        case S_MSG_AD_EMPTY:
            if (is_eot) {
                state = is_full ? S_TAG_INIT : S_MSG;
                is_partial = !is_full;
            } else {
                state = S_MSG;
            }
            break;
        case S_MSG:
            if (!is_partial) {
                if (is_eot) {
                    state = is_full ? S_TAG : S_MSG;
                    is_partial = !is_full;
                }
            } else {
                state = S_TAG_INIT_PARTIAL;
                is_partial = 0;
            }
            break;
        case S_TAG_INIT_PARTIAL:
        case S_TAG_INIT:
            state = S_TAG;
        default:
            break;
        }
    }
}
